@@ -2,6 +2,7 @@
 /*  $Header$
  *
  *  Copyright (c) 1987-2008 Sun Microsystems, Inc. All Rights Reserved.
+ *  Copyright (c) 2008 Robert Ancell
  *           
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -478,7 +479,7 @@ reset_display(void)
 
     ans = display_get_answer(&v->display);
     MPstr_to_num("0", DEC, ans);
-    display_clear(&v->display, FALSE);
+    display_clear(&v->display);
     display_set_number(&v->display, ans);
 }
 
@@ -490,8 +491,6 @@ ui_set_accuracy(int accuracy)
     char text[MAXLINE];
     char *desc, *current, *tooltip;
     
-    v->accuracy = accuracy;
-    
     SNPRINTF(text, MAXLINE, _("_Other (%d) ..."), accuracy);
     widget = gtk_bin_get_child(GTK_BIN(GET_WIDGET("acc_item_other")));
     gtk_label_set_markup_with_mnemonic(GTK_LABEL(widget), text);
@@ -524,7 +523,8 @@ ui_set_accuracy(int accuracy)
     set_int_resource(R_ACCURACY, accuracy);
     
     ui_make_registers();
-    display_refresh(&v->display, -1);
+    display_set_cursor(&v->display, -1);
+    display_refresh(&v->display);
     
     /* Hide the manual dialog */
     gtk_widget_hide(X->spframe);
@@ -613,7 +613,9 @@ ui_set_show_thousands_separator(gboolean visible)
     menu = GET_WIDGET("show_thousands_separator_menu");
     gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu), visible);
 
-    syntaxdep_show_display();
+    display_set_cursor(&v->display, -1);
+    display_refresh(&v->display);
+
     ui_make_registers();
 }
 
@@ -655,7 +657,9 @@ ui_set_show_trailing_zeroes(gboolean visible)
     v->show_zeroes = visible;
     set_boolean_resource(R_ZEROES, visible);
 
-    syntaxdep_show_display();
+    display_set_cursor(&v->display, -1);
+    display_refresh(&v->display);    
+
     ui_make_registers();
 
     menu = GET_WIDGET("show_trailing_zeroes_menu");
@@ -710,6 +714,70 @@ make_hostname()
 }
 
 
+gchar *
+ui_get_display(void)
+{
+    GtkTextIter start, end;
+    gtk_text_buffer_get_bounds(X->display_buffer, &start, &end);
+    return (gtk_text_buffer_get_text(X->display_buffer,
+                                     &start,
+                                     &end,
+                                     FALSE));
+}
+
+
+static int
+get_cursor(void)
+{
+    gint pos;
+    g_object_get(G_OBJECT(X->display_buffer), "cursor-position", &pos, NULL);
+    
+    /* Convert the last position to -1 */
+    if (pos == gtk_text_buffer_get_char_count(X->display_buffer)) {
+        return (-1);
+    } else {
+        return (pos);
+    }
+}
+
+
+static void
+set_bit_panel(void)
+{
+    int bit_str_len, i;
+    int MP[MP_SIZE];
+    char bit_str[MAXLINE], label[MAXLINE];
+
+    if (display_is_usable_number(&v->display, MP) || !is_integer(MP)) {
+        gtk_widget_set_sensitive(X->bit_panel, FALSE);
+        return;
+    }
+    make_fixed(bit_str, MAXLINE, MP, BIN, MAXLINE, FALSE);
+    bit_str_len = strlen(bit_str);
+    if (bit_str_len <= MAXBITS) {
+        gtk_widget_set_sensitive(X->bit_panel, TRUE);
+        
+        for (i = 0; i < MAXBITS; i++) {
+            if (i < bit_str_len) {
+                SNPRINTF(label, MAXLINE, " %c", bit_str[bit_str_len-i-1]);
+            } else {
+                SNPRINTF(label, MAXLINE, " 0");
+            }
+            gtk_label_set_text(GTK_LABEL(X->bits[MAXBITS - i - 1]), label);
+        }
+    } else {
+        gtk_widget_set_sensitive(X->bit_panel, FALSE);
+    }
+}
+
+
+static void do_button(int function, int arg)
+{
+    do_expression(function, arg, get_cursor());
+    set_bit_panel();
+}
+
+
 void
 ui_set_mode(enum mode_type mode)
 {
@@ -723,13 +791,13 @@ ui_set_mode(enum mode_type mode)
 
         ui_set_base(DEC);
         ui_set_numeric_mode(FIX);
-        ui_set_accuracy(DEFAULT_ACCURACY);
+        do_button(KEY_SET_ACCURACY, DEFAULT_ACCURACY);
         ui_set_show_thousands_separator(FALSE);
         ui_set_show_trailing_zeroes(FALSE);
         ui_make_registers();
 
         /* Reset display */
-        do_clear();
+        display_reset(&v->display);
         ui_set_statusbar("", "");
     }
     
@@ -833,34 +901,6 @@ ui_set_statusbar(const gchar *text, const gchar *imagename)
     gtk_statusbar_push(GTK_STATUSBAR(X->statusbar), 0, text); 
 }
 
-static void
-set_bit_panel(void)
-{
-    int bit_str_len, i;
-    int MP[MP_SIZE];
-    char bit_str[MAXLINE], label[MAXLINE];
-
-    if (display_is_usable_number(&v->display, MP) || !is_integer(MP)) {
-        gtk_widget_set_sensitive(X->bit_panel, FALSE);
-        return;
-    }
-    make_fixed(bit_str, MAXLINE, MP, BIN, MAXLINE, FALSE);
-    bit_str_len = strlen(bit_str);
-    if (bit_str_len <= MAXBITS) {
-        gtk_widget_set_sensitive(X->bit_panel, TRUE);
-        
-        for (i = 0; i < MAXBITS; i++) {
-            if (i < bit_str_len) {
-                SNPRINTF(label, MAXLINE, " %c", bit_str[bit_str_len-i-1]);
-            } else {
-                SNPRINTF(label, MAXLINE, " 0");
-            }
-            gtk_label_set_text(GTK_LABEL(X->bits[MAXBITS - i - 1]), label);
-        }
-    } else {
-        gtk_widget_set_sensitive(X->bit_panel, FALSE);
-    }
-}
 
 static gboolean
 redo_display(gpointer data)
@@ -1025,33 +1065,6 @@ ui_set_registers_visible(gboolean visible)
 }
 
 
-gchar *
-ui_get_display(void)
-{
-    GtkTextIter start, end;
-    gtk_text_buffer_get_bounds(X->display_buffer, &start, &end);
-    return (gtk_text_buffer_get_text(X->display_buffer,
-                                     &start,
-                                     &end,
-                                     FALSE));
-}
-
-
-static int
-get_cursor(void)
-{
-    gint pos;
-    g_object_get(G_OBJECT(X->display_buffer), "cursor-position", &pos, NULL);
-    
-    /* Convert the last position to -1 */
-    if (pos == gtk_text_buffer_get_char_count(X->display_buffer)) {
-        return (-1);
-    } else {
-        return (pos);
-    }
-}
-
-
 /*ARGSUSED*/
 static void
 about_cb(GtkWidget *widget)
@@ -1108,7 +1121,6 @@ cell_edited_cb(GtkCellRendererText *cell, const gchar *path_string,
     GtkTreeModel *model = (GtkTreeModel *) data;
     GtkTreePath *path = gtk_tree_path_new_from_string(path_string);
     GtkTreeIter iter;
-    gchar *old_text;
     gint *column;
 
     column = g_object_get_data(G_OBJECT(cell), "column");
@@ -1117,17 +1129,13 @@ cell_edited_cb(GtkCellRendererText *cell, const gchar *path_string,
 
     switch (GPOINTER_TO_INT(column)) {
         case COLUMN_VALUE:
-            gtk_tree_model_get(model, &iter, column, &old_text, -1);
-            g_free(old_text); 
             gtk_list_store_set(GTK_LIST_STORE(model), &iter, column,
-                           g_strdup(new_text), -1);
+                               g_strdup(new_text), -1);
             break;
 
         case COLUMN_DESCRIPTION:
-            gtk_tree_model_get(model, &iter, column, &old_text, -1);
-            g_free(old_text);
             gtk_list_store_set(GTK_LIST_STORE(model), &iter, column,
-                     g_strdup(new_text), -1);
+                               g_strdup(new_text), -1);
             break;
     }
  
@@ -1216,7 +1224,7 @@ void
 disp_cb(GtkWidget *widget)
 {
     if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)))
-        do_numtype((enum num_type) g_object_get_data(G_OBJECT(widget), "numeric_mode"));
+        do_button(KEY_SET_NUMBERTYPE, (int)g_object_get_data(G_OBJECT(widget), "numeric_mode"));
 }
 
 
@@ -1229,18 +1237,11 @@ base_cb(GtkWidget *widget)
     base = (enum base_type) g_object_get_data(G_OBJECT(widget),
                                               "base_mode");
     if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))) {
-        do_base(base);
+        do_button(KEY_SET_BASE, base);
     }
 }
 
 
-static void do_button(int function, int arg)
-{
-    do_expression(function, arg, get_cursor());
-    set_bit_panel();
-}
-
-
 static void
 help_display(void)
 {
@@ -1499,8 +1500,8 @@ create_constants_model()
         gtk_list_store_set(model, &iter,
                            COLUMN_NUMBER, i,
                            COLUMN_EDITABLE, TRUE,
-                           COLUMN_VALUE, g_strdup(constant),
-                           COLUMN_DESCRIPTION, g_strdup(v->con_names[i]),
+                           COLUMN_VALUE, constant,
+                           COLUMN_DESCRIPTION, v->con_names[i],
                            -1);
     }
 
@@ -1524,8 +1525,8 @@ create_functions_model()
         gtk_list_store_set(model, &iter,
                            COLUMN_NUMBER, i,
                            COLUMN_EDITABLE, TRUE,
-                           COLUMN_VALUE, g_strdup(v->fun_vals[i]),
-                           COLUMN_DESCRIPTION, g_strdup(v->fun_names[i]),
+                           COLUMN_VALUE, v->fun_vals[i],
+                           COLUMN_DESCRIPTION, v->fun_names[i],
                            -1);
     }
 
@@ -1645,8 +1646,8 @@ bit_toggle_cb(GtkWidget *event_box, GdkEventButton *event)
     number = (double) lval;
 
     mp_set_from_double(number, display_get_answer(&v->display));
-    display_set_string(&v->display, "Ans");
-    display_refresh(&v->display, -1);
+    display_set_string(&v->display, "Ans", -1);
+    display_refresh(&v->display);
 
     return (TRUE);
 }
@@ -1725,11 +1726,14 @@ get_constant(int n)
  
     SNPRINTF(vkey, MAXLINE, "constant%1dvalue", n);
     if ((vline = get_resource(vkey)) == NULL) {
+        g_free(nline);
         return;
     }   
 
     MPstr_to_num(vline, DEC, v->MPcon_vals[n]);
     STRNCPY(v->con_names[n], nline, MAXLINE - 1);
+    g_free(nline);
+    g_free(vline);
 }
 
 
@@ -1763,16 +1767,19 @@ get_function(int n)
  
     SNPRINTF(nkey, MAXLINE, "function%1dname", n);
     if ((nline = get_resource(nkey)) == NULL) {
-        return;  
+        return;
     }    
  
     SNPRINTF(vkey, MAXLINE, "function%1dvalue", n);
     if ((vline = get_resource(vkey)) == NULL) {
+        g_free(nline);
         return;
     }   
  
     STRNCPY(v->fun_vals[n], convert(vline), MAXLINE - 1);
     STRNCPY(v->fun_names[n], nline, MAXLINE - 1);
+    g_free(nline);
+    g_free(vline);
 }
 
 
@@ -1898,34 +1905,34 @@ kframe_key_press_cb(GtkWidget *widget, GdkEventKey *event)
     if (state == GDK_CONTROL_MASK && v->modetype == SCIENTIFIC) {
         switch (event->keyval) {
             case GDK_0:
-                do_accuracy(0);
+                do_button(KEY_SET_ACCURACY, 0);
                 return (TRUE);
             case GDK_1:
-                do_accuracy(1);
+                do_button(KEY_SET_ACCURACY, 1);
                 return (TRUE);
             case GDK_2:
-                do_accuracy(2);
+                do_button(KEY_SET_ACCURACY, 2);
                 return (TRUE);
             case GDK_3:
-                do_accuracy(3);
+                do_button(KEY_SET_ACCURACY, 3);
                 return (TRUE);
             case GDK_4:
-                do_accuracy(4);
+                do_button(KEY_SET_ACCURACY, 4);
                 return (TRUE);
             case GDK_5:
-                do_accuracy(5);
+                do_button(KEY_SET_ACCURACY, 5);
                 return (TRUE);
             case GDK_6:
-                do_accuracy(6);
+                do_button(KEY_SET_ACCURACY, 6);
                 return (TRUE);
             case GDK_7:
-                do_accuracy(7);
+                do_button(KEY_SET_ACCURACY, 7);
                 return (TRUE);
             case GDK_8:
-                do_accuracy(8);
+                do_button(KEY_SET_ACCURACY, 8);
                 return (TRUE);
             case GDK_9:
-                do_accuracy(9);
+                do_button(KEY_SET_ACCURACY, 9);
                 return (TRUE);
         }
     }
@@ -2063,8 +2070,10 @@ get_proc(GtkClipboard *clipboard, const gchar *buffer, gpointer data)
     }
     *dstp++ = '\0';
 
-    display_insert(&v->display, (char *) text, get_cursor()); // FIXME: Move out of gtk.c
-    display_refresh(&v->display, -1);
+    display_set_cursor(&v->display, get_cursor()); // FIXME: Move out of gtk.c
+    display_insert(&v->display, (char *) text);
+    display_set_cursor(&v->display, -1);
+    display_refresh(&v->display);
     free(text);
 }
 
@@ -2103,7 +2112,7 @@ popup_paste_cb(GtkWidget *menu)
 static void
 undo_cb(GtkWidget *widget)
 {
-    perform_undo();
+    do_button(KEY_UNDO, 0);
 }
 
 
@@ -2111,8 +2120,9 @@ undo_cb(GtkWidget *widget)
 static void
 redo_cb(GtkWidget *widget)
 {
-    perform_redo();
-    display_refresh(&v->display, -1);
+    do_button(KEY_REDO, 0);    
+    display_set_cursor(&v->display, -1);
+    display_refresh(&v->display);
 }
 
 
@@ -2232,7 +2242,7 @@ accuracy_radio_cb(GtkWidget *widget)
     int count;
     count = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(widget), "accuracy"));
     if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(widget))) {
-        do_accuracy(count);
+        do_button(KEY_SET_ACCURACY, count);
     }
 }
 
@@ -2253,7 +2263,7 @@ accuracy_other_cb(GtkWidget *widget)
 static void
 accuracy_default_cb(GtkWidget *widget)
 {
-    do_accuracy(DEFAULT_ACCURACY);
+    do_button(KEY_SET_ACCURACY, DEFAULT_ACCURACY);
 }
 
 
@@ -2283,7 +2293,7 @@ spframe_response_cb(GtkWidget *dialog, gint response_id)
     int val;
     if (response_id == GTK_RESPONSE_OK) {
         val = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(X->precision_spin));
-        ui_set_accuracy(val);
+        do_button(KEY_SET_ACCURACY, val);
     }
     
     gtk_widget_hide(dialog);
@@ -2705,6 +2715,8 @@ read_cfdefs()        /* Read constant/function definitions. */
 void
 ui_init(int *argc, char ***argv)
 {  
+    gchar *path;
+    
     X = (XVars) LINT_CAST(calloc(1, sizeof(struct Xobject)));
         
     gtk_init(argc, argv);
@@ -2714,7 +2726,9 @@ ui_init(int *argc, char ***argv)
     gtk_rc_get_default_files();
 
     v->home = (char *) g_get_home_dir();
-    gtk_rc_parse(g_build_path(v->home, RCNAME, NULL));
+    path = g_build_path(v->home, RCNAME, NULL);
+    gtk_rc_parse(path);
+    g_free(path);
 
     gtk_window_set_default_icon_name("accessories-calculator");
 }
@@ -2735,7 +2749,12 @@ ui_load(void)
     
     /* Load configuration */
     resource = get_resource(R_BITCALC);
-    show_bit = resource != NULL && strcmp(resource, Rcstr[0]) != 0;
+    if(resource) {
+        show_bit = strcmp(resource, Rcstr[0]) != 0;
+        g_free(resource);
+    } else {
+        show_bit = FALSE;
+    }
 
     ui_set_show_thousands_separator(v->show_tsep);
     ui_set_show_trailing_zeroes(v->show_zeroes);
@@ -2779,9 +2798,6 @@ ui_start(void)
 
     gtk_widget_show(X->kframe);
 
-    /* Init expression mode.
-     * This must be executed after do_base is called at init.
-     */    
     reset_display();
 
     gtk_main();
